class Constraint():
    def __init__(self):
        pass

    def get_constraint(self):
        print("get_constraint method is not implemented.")

    def do_eval(self, q):
        print("do_eval method is not implemented.")
